/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 ***************************************************************************/

#include "diagnostic.h"
#include "io.h"
#include "args.h"
#include "symbol_table.h"
#include "chew.h"
#include "if_control.h"
#include "line_despatch.h"
#include "evaluator.h"
#include "opts.h"

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
 
/*!\ingroup diagnostics_module, diagnostics_interface, diagnostics_internals
 *\file diagnostic.c
 * This file implements the Diagnostics module
 */

/*! \addtogroup diagnostics_internals */
/*@{*/

/*! Prefixes for debugging names of the comment states */
static char const * const comment_name[] = {
	"NO","C","CXX",	"STARTING",	"FINISHING","PSEUDO"
};

/*! Prefixes for debugging names of the line states */
static char const * const linestate_name[] = {
	"VACANT","DIRECTIVE","CODE"
};

/*! Debugging names of the \c #if states */
static char const * const ifstate_name[] = {
	"OUTSIDE", "FALSE_PREFIX", "TRUE_PREFIX",
	"PASS_MIDDLE", "FALSE_MIDDLE", "TRUE_MIDDLE",
	"PASS_ELSE", "FALSE_ELSE", "TRUE_ELSE",
	"FALSE_TRAILER"
};

/*! Debugging names of the input line types */
static char const * const linetype_name[] = {
	"TRUEI", "FALSEI", "IF", "TRUE", "FALSE",
	"ELIF", "ELTRUE", "ELFALSE", "ELSE", "ENDIF",
	"DODGY TRUEI", "DODGY FALSEI",
	"DODGY IF", "DODGY TRUE", "DODGY FALSE",
	"DODGY ELIF", "DODGY ELTRUE", "DODGY ELFALSE",
	"DODGY ELSE", "DODGY ENDIF",
	"PLAIN", "EOF",
	"CONSISTENT DEFINE KEEP","CONSISTENT DEFINE DROP",
	"CONTRADICTORY DEFINE","DIFFERING DEFINE",
	"CONSISTENT UNDEF KEEP", "CONSISTENT UNDEF DROP",
	"CONTRADICTORY UNDEF"
};


/*! Get the debugging name of the current comment state
 *	\return A pointer to the name in static storage.
 */
static char const *
get_comment_type_name(void) 
{
	return comment_name[GET_PUBLIC(chew,comment_state)];
}

/*! Get the debugging name of the current line state
 *	\return A pointer to the name in static storage.
 */
static char const *
get_linestate_name(void)
{
	return linestate_name[GET_PUBLIC(chew,line_state)];
}

/*! Get the debugging name of the current \c #if state
 *	\return A pointer to the name in static storage.
 */
static char const *
if_state_name(void)
{
	return ifstate_name[if_state()];
}

/*! Get the debugging name of a line type.
 *
 *	\param	type	The line type whose name is wanted.
 *	\return	The debugging name of \em type.
 */
static char const *
get_linetype_name(line_type_t type)
{ 
	return linetype_name[type];
}


/*!	Produce a diagnostic message.
 *  The function interface is the same as that of \c complain(), with
 *	the exception that the parameter <tt>va_list argp</tt> is a
 *	pointer that the caller has initialised with \c va_start() to
 *	address a list of unspecified arguments.
 *
 *	If the function is called with <tt>buf == NULL</tt>, then it
 *	must provide its own buffer in which to compose a diagnostic.
 *	But this buffer cannot be allocated on the heap, since the
 *	problem to be reported may be exhaustion of the heap. In this
 *	case therefore the function allocates an automatic buffer on
 *  the stack and then calls itself twice: once with arguments to
 * 	that direct it to compose the diagnostic in this buffer and
 *	then again with arguments to write out the diagnostic stored
 *	there.
 *
 *	All diagnostic messages are ultimately composed and output by this function. 
 */
static void
vcomplain(reason_code_t reason, char *buf, const char *format, va_list argp);

/*@}*/


/*! \ingroup diagnostics_internals_state_utils */
/*@{*/

/*! The global state of the Diagnostic module */ 
STATE_DEF(diagnostic) {
	bool		dbg; /*!< Is --debug enabled? */
	unsigned int	exitstat; /*!< Program exit status */
} STATE_T(diagnostic);
/*@}*/

/*! \addtogroup diagnostics_internals_state_utils */
/*@{*/
NO_PUBLIC_STATE(diagnostic);

IMPLEMENT(diagnostic,ZERO_INITABLE);
/*@}*/

static void
vcomplain(reason_code_t reason, char *buf, const char *format, va_list argp)
{
	if (GET_PUBLIC(args,done_cmdline) && reason != SEVERITY_NONE) {
		int i;
		int masks = GET_PUBLIC(args,masks);
		for ( i = 0; i < masks; ++i) {
			int mask = GET_PUBLIC(args,suppression_masks)[i];
			if ((reason & mask) == mask) {
				SET_STATE(diagnostic,exitstat) |= reason & SEVERITY_MASK;
				return;
			}
		}
	}
	if (buf) {
		if (format) {
			char *bufp = buf;
			size_t depth = if_depth();
			sprintf(bufp,"%s: ",GET_PUBLIC(args,prog_name));
			bufp += strlen(bufp);
			if (GET_PUBLIC(io,filename)) {
				sprintf(bufp,"%s: ",GET_PUBLIC(io,filename));
				bufp += strlen(bufp);
			}
			if (GET_PUBLIC(io,line_num)) {
				sprintf(bufp,"line %d: ",(int)GET_PUBLIC(io,line_num));
				bufp += strlen(bufp);
			}
			if (reason & ISSUE_MASK) {
				char * category;
				switch(reason & SEVERITY_MASK) {
				case SEVERITY_INFO:
					category = "info";
					break;
				case SEVERITY_WARNING:
					category = "warning";
					break;
				case SEVERITY_ERROR:
					category = "error";
					break;
				case SEVERITY_ABEND:
					category = "abend";
					break;
				default:
					assert(false);
				}
				sprintf(bufp,"%s 0x%04x: ",category,reason);
				bufp += strlen(bufp);
				SET_STATE(diagnostic,exitstat) |= reason & SEVERITY_MASK;
			}
			vsprintf(bufp,format,argp);
			bufp += strlen(bufp);
			if (depth > 0) {
				sprintf(bufp," (#if line %d depth %d)\n",
					(int)if_start_line(),(int)depth);
			}
			else {
				bufp[0] ='\n';
				bufp[1] = '\0';
			}
		}
		else {
			fputs(buf,stderr);
		}
	}
	else {
		char buffer[GRIPE_MAXLINE];
		vcomplain(reason,buffer,format,argp);
		vcomplain(reason,buffer,NULL,NULL);
	}
}


/* API ***************************************************************/

size_t
safe_len(char const *str)
{
	size_t len = strlen(str);
	if (str[len - 1] == '\n') {
		--len;
	}
	if (len > MAX_QUOTE) {
		len = MAX_QUOTE;
	}
	return len;
}

int
exitcode(void)
{
	int severities;
	int summaries;
	int ret = GET_STATE(diagnostic,exitstat);
	if (GET_PUBLIC(line_despatch,lines_dropped)) {
		ret |= SUMMARY_DROPPED_LINES;
	}
	if (GET_PUBLIC(line_despatch,lines_changed)) {
		ret |= SUMMARY_CHANGED_LINES;
	}
	SET_STATE(diagnostic,exitstat) = ret;
	severities = ret >> SEVERITY_SHIFT;
	summaries = (ret & SUMMARY_MASK) << GRIPE_SHIFT;
	return severities | summaries;
}

int
get_exit_flags(int bits)
{
	return bits &
		(SUMMARY_MASK | SEVERITY_MASK) & GET_STATE(diagnostic,exitstat);
}

void
set_exit_flags(int bits, bool set)
{
	bits &= (SUMMARY_MASK | SEVERITY_MASK);
	if (set) {
		SET_STATE(diagnostic,exitstat) |= bits;
	}
	else {
		SET_STATE(diagnostic,exitstat) &= ~bits;
	}
}

void
complain(reason_code_t reason, char *buf, const char *format,...)
{
	va_list argp;
	va_start(argp,format);
	vcomplain(reason,buf,format,argp);
	va_end(argp);
}

void
parse_error(int reason, const char *format,...)
{
	va_list argp;
	va_start(argp,format);
	vbail(reason,format,argp);
	va_end(argp);
}

void
bad_args_error(char const *sym, char const *symend, int symind)
{
	static char * formats[] = {
		"Duplicated argument '-%c%s'",
		"Argument '-%c%s' contradicts previous '-%c%s=%s'",
		"Argument '-%c%s' contradicts previous '-%c%s'"
	};
	
	enum {
		DUPLICATE,
		CONTRADICTS_NAME_EQ_VAL,
		CONTRADICTS_NAME
	};
		
	char const *this_val = 0;
	eval_result_t * symbol = SYMBOL(symind);
	char const *prev_val = symbol->sym_def;
	char const *name = symbol->sym_name;
	char this_opt = sym[-1];
	char prev_opt = prev_val ? 'D' : 'U';
	if (this_opt == 'D') {
		if (*symend == '=') {
			this_val = ++symend;
		}
		else {
			this_val = "";
		}
	}
	if (this_opt == prev_opt) {
		if (this_opt == 'D') {
			if (!strcmp(this_val,prev_val)) {
				complain(	GRIPE_DUPLICATE_ARGS,
							NULL,
							formats[DUPLICATE],
							this_opt,
							sym);
			}
			else if (prev_val[0]) {
				bail(GRIPE_INVALID_ARGS,formats[CONTRADICTS_NAME_EQ_VAL],
					this_opt,sym,prev_opt,name,prev_val);
			}
			else {
				bail(GRIPE_INVALID_ARGS,formats[CONTRADICTS_NAME],
					this_opt,sym,prev_opt,name);
			}
		}
		else { /* Both -U, so can only be duplicate */
			complain(GRIPE_DUPLICATE_ARGS,NULL,formats[DUPLICATE],this_opt,sym);
		}
	}
	else if (this_opt == 'D') { /* Previous must be -U */
		bail(GRIPE_INVALID_ARGS,formats[CONTRADICTS_NAME],
				this_opt,sym,prev_opt,name);
	}
	else if (prev_val[0]) { /* This = -UNAME, previous = -DNAME=STRING */
			bail(GRIPE_INVALID_ARGS,formats[CONTRADICTS_NAME_EQ_VAL],
				this_opt,sym,prev_opt,name,prev_val);
	}
	else { /* This = -UNAME, previous = -DNAME */
		bail(GRIPE_INVALID_ARGS,formats[CONTRADICTS_NAME],
				this_opt,sym,prev_opt,name);
	}
}

void
give_up_confused(void)
{
	bail(GRIPE_CONFUSED,"Bailing out in confusion");
}

/*! Exit the program with a specified return code and diagnostic
 *	message.
 *  The function interface is the same as that of \c bail(), with
 *	the exception that the parameter <tt>va_list argp</tt> is a
 *	pointer that the caller has initialised with \c va_start() to
 *	address a list of unspecified arguments.
 *
 *	All functions that terminate the program do so by calling this function 
 */
void
vbail(int reason, const char *msg, va_list argp)
{
	if (msg) {
		vcomplain(reason,NULL,msg,argp);
	}
	else {
		SET_STATE(diagnostic,exitstat) |= reason & SEVERITY_MASK;
	}
	exit(exitcode());
}


void
bail(int reason, const char *format,...)
{
	va_list argp;
	va_start(argp,format);
	vbail(reason,format,argp);
	va_end(argp);
}

void
debugging(bool on)
{
	SET_STATE(diagnostic,dbg) = on;
}

void 
debug(dbg_code_t how,...)
{
	if (GET_STATE(diagnostic,dbg)) {
		va_list ap;
		va_start(ap, how);
		switch(how)
		{
		case DBG_1:
			vcomplain(SEVERITY_NONE,NULL,"eval%d !",ap);
			break;
		case DBG_2:
			vcomplain(SEVERITY_NONE,NULL,"eval%d (",ap);
			break;
		case DBG_3:
			vcomplain(SEVERITY_NONE,NULL,"eval%d number",ap);
			break;
		case DBG_4:
			vcomplain(SEVERITY_NONE,NULL,"eval%d defined",ap);
			break;
		case DBG_5:
			vcomplain(SEVERITY_NONE,NULL,"eval%d symbol",ap);
			break;
		case DBG_6:
			vcomplain(SEVERITY_NONE,NULL,"eval%d bad expr",ap);
			break;
		case DBG_7:
			vcomplain(SEVERITY_NONE,NULL,"eval%d = %d",ap);
			break;
		case DBG_8:
			vcomplain(SEVERITY_NONE,NULL,"eval%d",ap);
			break;
		case DBG_9:
			vcomplain(SEVERITY_NONE,NULL,"eval%d %s",ap);
			break;
		case DBG_10:
			vcomplain(SEVERITY_NONE,NULL,"eval%d = %d",ap);
			break;
		case DBG_11:
			vcomplain(SEVERITY_NONE,NULL,"eval %s",ap);
			break;
		case DBG_12:
			vcomplain(SEVERITY_NONE,NULL,"eval = %d",ap);
			break;
		case DBG_13:
			vcomplain(SEVERITY_NONE,NULL,"eval #define %.*s",ap);
			break;
		case DBG_14:
		{
			size_t safelen = va_arg(ap,size_t);
			char const * define = va_arg(ap,char const *);
			int linetype = va_arg(ap,int);
			char const * const linetype_name = get_linetype_name(linetype);
			complain(SEVERITY_NONE,NULL,"eval #define %.*s = %s",
				safelen,define,linetype_name);
			break;
		}
		case DBG_15:
			vcomplain(SEVERITY_NONE,NULL,"eval #undef %.*s",ap);
			break;
		case DBG_16:
		{
			size_t safelen = va_arg(ap,size_t);
			char const * define = va_arg(ap,char const *);
			int linetype = va_arg(ap,int);
			char const * const linetype_name = get_linetype_name(linetype);
			complain(SEVERITY_NONE,NULL,"eval #undef %.*s = %s",
				safelen,define,linetype_name);
			break;
		}
		case DBG_17:
			complain(SEVERITY_NONE,NULL,"parser %s comment %s line",
				get_comment_type_name(), get_linestate_name());
			break;
		case DBG_18:
			vcomplain(SEVERITY_NONE,NULL,"find_sym %s %s",ap);
			break;
		case DBG_19:
		{
			int linetype = va_arg(ap,int);
			complain(SEVERITY_NONE,NULL,"process %s -> %s depth %d",
		    	get_linetype_name(linetype),
		    	if_state_name(),
				if_depth());
			break;
		}
		case DBG_20:
			vcomplain(SEVERITY_NONE,NULL,"eval%d +",ap);
			break;
		case DBG_21:
			vcomplain(SEVERITY_NONE,NULL,"eval%d -",ap);
			break;
		case DBG_22:
			vcomplain(SEVERITY_NONE,NULL,"eval%d ~",ap);
			break;
		default:
			assert(false);
		}
		va_end(ap);
	}
}

void
exit_diagnostics(void)
{
	int ret = exitcode();
	SET_PUBLIC(io,line_num) = 0;

	if (GET_STATE(diagnostic,exitstat) & SEVERITY_ABEND) {
		complain(SEVERITY_NONE,NULL,"Aborted, exit code 0x%04x",ret);
	}
	else {
		char * eof_status = (input_opened() && input_eof()) ? "at" : "before";
		char * diagnostic_status = "";

		if (GET_STATE(diagnostic,exitstat) & SEVERITY_ERROR) {
			diagnostic_status = " with errors";
		}
		else if (GET_STATE(diagnostic,exitstat) & SEVERITY_WARNING) {
			diagnostic_status = " with warnings";
		}
		else if (GET_STATE(diagnostic,exitstat) & SEVERITY_INFO) {
			diagnostic_status = " with remarks";
		}
		if (!input_opened()) {
			complain(SEVERITY_NONE,NULL,
				"Completed%s, exit code 0x%04x",
					diagnostic_status,ret);
		}
		else {
			complain(SEVERITY_INFO,NULL,
				"Completed %s EOF%s, exit code 0x%04x",
				eof_status,
				diagnostic_status,
				ret);
			if (GET_STATE(diagnostic,exitstat) &
				SUMMARY_DROPPED_LINES & ISSUE_MASK) {
				complain(SUMMARY_DROPPED_LINES,NULL,"Input lines were dropped");
			}
			if (GET_STATE(diagnostic,exitstat) &
				SUMMARY_CHANGED_LINES & ISSUE_MASK) {
				complain(SUMMARY_CHANGED_LINES,NULL,"Input lines were changed");
			}	
			if (GET_STATE(diagnostic,exitstat) &
				SUMMARY_ERRORED_LINES & ISSUE_MASK) {
				complain(	SUMMARY_ERRORED_LINES,
							NULL,
							"Input lines were changed to #error directives");
			}
			if (GET_STATE(diagnostic,exitstat) &
				SUMMARY_ERROR_OUTPUT & ISSUE_MASK) {
				complain(	SUMMARY_ERRORED_LINES,
							NULL,
							"Unconditional #error directives were output");
			}
		}
	}
}

void 
orphan_elif(void)
{
	parse_error(GRIPE_ORPHAN_ELIF,"Orphan #elif");
}

void
orphan_else(void)
{
	parse_error(GRIPE_ORPHAN_ELSE,"Orphan #else");
}

void
orphan_endif(void)
{
	parse_error(GRIPE_ORPHAN_ENDIF,"Orphan #endif");
}

void
early_eof(void)
{
	parse_error(GRIPE_EOF_TOO_SOON,"Unexpected EOF");
}

void
obfuscation(void)
{
	parse_error(GRIPE_OBFUSC_DIRECTIVE,
				"Obfuscated preprocessor control line");
}


/* EOF */


