#ifndef DIAGNOSTIC_H
#define DIAGNOSTIC_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 ***************************************************************************/

#include "opts.h"
#include "bool.h"
#include <stddef.h>
#include <stdarg.h>
 
 /*!\ingroup diagnostics_module diagnostics_interface
 *  \file diagnostic.h
 *	This file provides the Diagnostics module interface
 */

/*!\addtogroup diagnostics_interface */
/*@{*/



#define GRIPE_MAXLINE	2048
	/*!< Max length of diagnostic we will insert on output */
	
#define MAX_QUOTE (GRIPE_MAXLINE / 2)	
	/*!< Max initial part of text we will quote in a diagnostic */
 
/*! Enumeration of all diagnostic reason codes and auxiliary bit masks and
 *	offsets.
 *	A complete reason code consists or an or-combination
 *	of a severity code and an issue code.
 *	A range of issue codes are \em summary codes, representing significant
 *	diagnostics accrued in the course of processing all inputs.
 *	Another range of issue codes are \em gripes, representing complaints
 *	specific to input fragments. 
 */
typedef enum {
	/*! The severity bitfield is offset this amount */
	SEVERITY_SHIFT = 10,
	/*! &-ed with code strips off severity */
	ISSUE_MASK = (1 << SEVERITY_SHIFT) - 1,
	/*! No severity */
	SEVERITY_NONE = 0 << SEVERITY_SHIFT,
	/*! Informational diagnositic */
	SEVERITY_INFO = 1 << SEVERITY_SHIFT,
	/*! Warning diagnostic */
	SEVERITY_WARNING = 2 << SEVERITY_SHIFT,
	/*! User error diagnostic */
	SEVERITY_ERROR = 4 << SEVERITY_SHIFT,
	/*! Fatal internal or environment error diagnostic */
	SEVERITY_ABEND = 8 << SEVERITY_SHIFT,
	/*! &-ed with code extracts severity */
	SEVERITY_MASK = 255 << SEVERITY_SHIFT,
	/*! Min value of summary diagnostic bitfield */
	SUMMARY_MIN = 1,
	/*! Summary diagnostic - input lines were dropped */
	SUMMARY_DROPPED_LINES = 1 | SEVERITY_INFO,
	/*! Summary diagnostic - input lines were changed */
	SUMMARY_CHANGED_LINES = 2 | SEVERITY_INFO,
	/*! Summary diagnostic - input lines were converted to #errors */
	SUMMARY_ERRORED_LINES = 4 | SEVERITY_WARNING,
	/*! Summary diagnostic - unconditional #error lines were output */
	SUMMARY_ERROR_OUTPUT = 8 | SEVERITY_WARNING,
	/*! All valid summary values are less than this */
	SUMMARY_MAX = 16,
	/*! &-ed with code extracts summary diagnostics */
	SUMMARY_MASK = SUMMARY_MAX - 1,
	/*! The gripes bitfield is offset this amount */
	GRIPE_SHIFT = 4,
	/*! Duplicate diagnoistic filters are specified with the \c --gag
	 * option
	 */
	GRIPE_DUPLICATE_MASK = (27 << GRIPE_SHIFT) | SEVERITY_INFO,
	/*! The same argument occurs for multiple \c --define or \c
	 *	--undefine options
	 */
	GRIPE_DUPLICATE_ARGS = (1 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! An input \c #def or \c #undef directive was deleted on
	 * 	output because it contradicts a \c --define or \c --undefine
	 *	option together with \c --contradict \c delete
	 */
	GRIPE_DELETED_CONTRADICTION = (2 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! An input \c #def or \c #undef directive was commented out on
	 *	output because it contradicts a \c --define or \c --undefine
	 *	option togther with \c --contradict \c comment
	 */
	GRIPE_COMMENTED_CONTRADICTION = (3 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! An input \c #def or \c #undef directive was converted to an
	 *	#error on output because it contradicts a \c --define or \c --undefine
	 *	option \c --contradict \c error
	 */
	GRIPE_ERRORED_CONTRADICTION = (4 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! An \c #error directive was input unconditionally, i.e.
	 *	not in the scope of any \c #if
	 */
	GRIPE_UNCONDITIONAL_ERROR_INPUT = (5 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! An \c #error directive was output unconditionally, i.e.
	 *	not in the scope of any \c #if
	 */
	GRIPE_UNCONDITIONAL_ERROR_OUTPUT = (6 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! Garbage text was input following a \c # directive
	 */
	GRIPE_GARBAGE_AFTER_DIRECTIVE = (7 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! More than \c MAXMASKS diagnostic filters are specified with the
	 *	\c --gag option
	 */
	GRIPE_TOO_MANY_MASKS = (28 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! The \c --verbose option is mixed with the \c --gag option */
	GRIPE_VERBOSE_ONLY = (29 << GRIPE_SHIFT) | SEVERITY_WARNING,
	/*! A divide by zero was found in an expression */
	GRIPE_ZERO_DIVIDE = (38 << GRIPE_SHIFT) | SEVERITY_WARNING, 
	/*! An orphan \c #elif was found in input */
	GRIPE_ORPHAN_ELIF = (8 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! An orphan \c #else was found in input */
	GRIPE_ORPHAN_ELSE = (9 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! An orphan \c #endif was found in input */
	GRIPE_ORPHAN_ENDIF = (10 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! Unexpected end of input */
	GRIPE_EOF_TOO_SOON = (11 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! An obfuscated \c # directive was found on input */
	GRIPE_OBFUSC_DIRECTIVE = (12 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! Input ends within a comment */
	GRIPE_EOF_IN_COMMENT = (13 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! Input ends within quotation */
	GRIPE_EOF_IN_QUOTE = (14 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! The commandline does not specify anything to do */
	GRIPE_NOTHING_TO_DO = (15 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! The commandline invalidly specifies multiple input files without
	 *	the \c --replace option
	 */	
	GRIPE_ONE_FILE_ONLY = (16 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! A newline occurs within quotation in input */
	GRIPE_NEWLINE_IN_QUOTE = (17 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! A newline is missing at end of input */
	GRIPE_MISSING_EOF_NEWLINE = (18 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! The commandline options are not a valid combination */
	GRIPE_INVALID_ARGS = (19 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! The commandline is syntactically invalid */
	GRIPE_USAGE_ERROR = (20 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! \c #if nesting exceeds \c MAXDEPTH in input */
	GRIPE_TOO_DEEP = (21 << GRIPE_SHIFT)  | SEVERITY_ERROR,
	/*! The \c --file option occurs more than once */
	GRIPE_MULTIPLE_ARGFILES = (31 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! The argument to a \c --define or \c --undefine option
	 *	contains mysterious characters.
	 */
	GRIPE_GARBAGE_ARG = (32 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! A valid preprocessor identifier was not found where required */
	GRIPE_NOT_IDENTIFIER = (33 << GRIPE_SHIFT) | SEVERITY_ERROR,
	/*! Cannot open an input file */
	GRIPE_CANT_OPEN_INPUT = (22 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Out of memory */
	GRIPE_OUT_OF_MEMORY = (23 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! A filename exceeds the maximum length for the environment  */
	GRIPE_FILENAME_TOO_LONG = (24 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! End of input encountered while reading a quoted filename.
		A closing quotation was not found */
	GRIPE_EOF_IN_FILENAME = (25 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Internal logic error */
	GRIPE_CONFUSED = (26 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Read error on input */
	GRIPE_CANT_READ_INPUT = (30 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Write error on output */
	GRIPE_FILE_WRITE_FAIL = (39 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/* MAX GRIPE = 40. When you add a new gripe, you must give it
		it the MAX GRIPE gripe number, increment MAX GRIPE in this
		comment and move this comment adjacent to your new gripe */ 
	/*! Invalid filename */
	GRIPE_ILLEGAL_FILENAME = (34 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Failure to delete a file */
	GRIPE_CANT_DELETE_FILE = (35 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Failure to rename a file */
	GRIPE_CANT_RENAME_FILE = (36 << GRIPE_SHIFT) | SEVERITY_ABEND,
	/*! Cannot create a temporary file */
	GRIPE_NO_TEMPFILE = (37 << GRIPE_SHIFT) | SEVERITY_ABEND

} reason_code_t;

/*! Arbitary codes to select the behaviour of the \c debug() function */
typedef enum {
	DBG_1 = 1,
	DBG_2,
	DBG_3,
	DBG_4,
	DBG_5,
	DBG_6,
	DBG_8,
	DBG_7,
	DBG_9,
	DBG_10,
	DBG_11,
	DBG_12,
	DBG_13,
	DBG_14,
	DBG_15,
	DBG_16,
	DBG_17,
	DBG_18,
	DBG_19,
	DBG_20,
	DBG_21,
	DBG_22
} dbg_code_t; 

/*! Set or clear severity or summary bits in the exit status.
 *	\param bits		The bits to be set or cleared
 *	\param set		If \em true, then \em bits are set in the exit status;
 *					otherwise they are cleared.
 */
extern void
set_exit_flags(int bits, bool set);

/*! Query the current exit status flags.
 *  \param bits		The exit status bits to be queried.
 *	\return A bit mask consisting of those bits that are set in \em bits
 *	and also in the exit status.
 */  
extern int
get_exit_flags(int bits);

/*! Convert the internal exit status to an exit code for \c exit(),
	manipulating the bits so that the exit diagnostic bits are
	returned in the low order byte.
	\return RC such that:
	
	- RC = 0:		SEVERITY_NONE: No diagnostics accrued.
	
	- RC & 0x1:		SEVERITY_INFO: 	At worst info diagnostics accrued.
					
	- RC & 0x2:		SEVERITY_WARNING: At worst warnings accrued.
				
	- RC & 0x4:		SEVERITY_ERROR:	At worst errors accrued.
				
	- RC & 0x8:		SEVERITY_ABEND:	Execution was aborted.
	
	- RC & 0x11:	SUMMARY_DROPPED_LINES (INFO): Input lines were dropped on
	 				output.
					
	- RC & 0x21:	SUMMARY_CHANGED_LINES (INFO): Input lines were changed on
					output.

	- RC & 0x42:	SUMMARY_ERRORED_LINES (WARNING): Lines were changed to
					\c #errors on output.
	
	- RC & 0x82:	SUMMARY_ERROR_OUPUT (WARNING): Unconditional \c #errors were
					output.
 */  
extern int exitcode(void);

/*! Get the length of a string excluding any final newline.
 *	\param str	The string to be measured.
 *	\return The length of \c str excluding any final newline.
 */ 
extern size_t
safe_len(char const *str);
 
/*!	Produce a diagnostic message.
 *	\param reason	The reason code of the diagnostic
 *	\param buf Either \c NULL or a pointer to a buffer in which the
 *	diagnostic may be composed or from which it may be written.
 *	\param format	Either \c NULL or a \c printf format from which the
 *	diagnostic may be composed together with the ensuring unspecified
 *	arguments.
 *
 *	The diagnostic is prefixed with the program name, and then
 *	the current input filename and line number, if any, that provoke the
 *	diagnostic.
 *
 *	Then follows one of:
 *
 *	- <tt>info \em reason</tt>
 *	- <tt>warning \em reason</tt>
 *	- <tt>error \em reason </tt>
 *	- <tt>abend \em reason </tt>
 *
 *	depending on the severity of \em reason.
 *
 *	The remainder of the diagnostic is composed from \em format and
 *	subsequent args.
 *
 *	The production of the diagnostic is determined as follows:
 *
 *	- If <tt>\em buf == 0 && \em format != 0</tt>, the diagnostic is
 *	is written to \c stderr.
 *	- If <tt>\em buf != 0 && \em format != 0</tt> the diagnostic is composed in
 *	\em buf but not written.
 *	- If <tt>\em buf != 0 && \em format == 0</tt> then then a diagnostic is presumed
 *	to have been composed in \em buf by an earlier call, and this diagnostic
 *	is written to \c stderr.
 *
 */
extern void
complain(reason_code_t reason, char *buf, const char *format,...);

/*! Exit the program on a fatal parsing error with a diagnostic message.
 *	\param reason	The reason code for the error.
 *	\param format	A \c printf format from which to compose the diagnostic
 *	together with subsequent arguments.
 */
extern void
parse_error(int reason, const char *format,...);

/*! Write a diagnostic on \c stderr when the same symbol occurs twice
 *	as an argument among the \c --define and \c --undefine options.
 *
 *	\param sym	The start of symbol that occurs twice.
 *	\param symend	The end of the symbol that occurs twice.
 *	\param symind The index of the symbol in the symbol table.
 * 
 *	If the argument occurs only for more than one \c --define or for more than
 *  one \c --undefine, then these options are merely duplicated and the
 *  diagnostic is a warning. 
 *
 *	If the argument occurs for a \c --define and also an \c --undefine,
 *	then these options are contradictory and the diagnostic is fatal.
 *
 *	If contradictory options are detected the function calls \c exit().
 */
extern void
bad_args_error(char const *sym, char const *symend, int symind);

/*! Exit the program on an internal logic error with an aplogetic message.
 *	The function is called if the program recognises that it
 *	it has lost its bearings.
 */
extern void
give_up_confused(void);

/*! Exit the program with a specified return code and diagnostic
 *	message.
 *  The function interface is the same as that of \c bail(), with
 *	the exception that the parameter <tt>va_list argp</tt> is a
 *	pointer that the caller has initialised with \c va_start() to
 *	address a list of unspecified arguments.
 *
 *	All functions that terminate the program do so by calling this function 
 */
extern void
vbail(int reason, const char *msg, va_list argp);


/*! Exit the program out with a specified return code and
 *	diagnostic message.
 *	\param retcode	The code to be returned by \c exit().
 *	\param format	A \c printf format from which to compose the diagnostic
 *	together with subsequent arguments.
 */
extern void
bail(int retcode, const char *format,...);

/*! Turn debugging output on or off
 *	\param on	If \em true, debugging output is enabled; otherwise disabled.
 */
extern void
debugging(bool on);

/*! Write debugging info on \c stderr.
 * \param how	A code that selects how the function will compose
 *	debugging output from the subsequent arguments.
 */
extern void 
debug(dbg_code_t how,...);

/*! Write summary diagnostics on \c stderr at exit.
 *	The summary diagnostics will be output unless \c --gag \c summary
 *	is in force, which it is by default.
 */
extern void
exit_diagnostics(void);

/*! Diagnose an input orphan\c #elif on \c stderr */ 
extern void 
orphan_elif(void);

/*! Diagnose an input orphan \c #else on \c stderr */ 
extern void
orphan_else(void);

/*! Diagnose an input orphan \c #endif on \c stderr */ 
extern void
orphan_endif(void);

/*! Diagnose unexpected end of input on \c stderr */ 
extern void
early_eof(void);

/*! Diagnose an input obfuscated \c # directive on \c stderr */ 
extern void
obfuscation(void);

/*@}*/

/*! \addtogroup diagnostics_interface_state_utils */
/*@{*/
IMPORT_INITOR(diagnostic);
IMPORT_FINITOR(diagnostic);
/*@}*/

#endif /* EOF */
